#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include <errno.h>

char *cftime(int f_time,int f_date,char *out_string);
void parse_attribs(struct ffblk *f_block);
char at_rdonly,at_hidden,at_system,at_label,at_direc,at_arch;

main(int num_parms,char **parms)
{
  struct ffblk file_block;
  struct dfree dfree_space;
  int flag,file_count,drive,done,index,col_count,num_cols,file_parm_flag,len;
  int zero,def_zero,num_switches;
  char attrib,*def_point,def_attrib,*point,character,switches[80];
  char name[9],pname[9],ext[5],pext[5],date_time_string[19],file_parm[80];
  char proot[8],pdrive[3],ppath[80];
  unsigned long int total_size,free_space;

  /* Read the ADDEFCOL environment variable to set the default number of   */
  /* display columns. If the ADDEFCOL is undefined or invalid default to 1 */

  def_point = getenv("ADDEFCOL");
  if (def_point) {
     switch (def_point[0]) {
	case '1' : num_cols = 1;break;
	case '2' : num_cols = 2;break;
	case '5' : num_cols = 5;break;
	default	 :
	   printf("Invalid ADDEFCOL environment variable %c\n",def_point[0]);
	   exit(1);
     }
  }
  else {
     num_cols = 1;
  }
  /* set the column count to zero */
  col_count = 0;

  /* Read in the ADDEFATT environment variable and parse the attribute	   */
  /* selection switches. If ADDEFATT is undefined, set the default to 0x3f */
  def_point = getenv("ADDEFATT");
  if (def_point) {
     len = strlen(def_point);
     def_attrib = 0;
     def_zero = 0;
     for (index = 0;index < len;index++) {
	character = def_point[index];
	point = strchr(" /AaDdHhLlRrSsXxOo0",character);
	if (point) {
	   switch (character) {
	      case 'A' :
	      case 'a' : def_attrib |= FA_ARCH;break;
	      case 'D' :
	      case 'd' : def_attrib |= FA_DIREC;break;
	      case 'H' :
	      case 'h' : def_attrib |= FA_HIDDEN;break;
	      case 'L' :
	      case 'l' : def_attrib |= FA_LABEL;break;
	      case 'R' :
	      case 'r' : def_attrib |= FA_RDONLY;break;
	      case 'S' :
	      case 's' : def_attrib |= FA_SYSTEM;break;
	      case 'X' :
	      case 'x' : def_attrib = 0x3f;break;
	      case 'O' :
	      case 'o' :
	      case '0' : def_zero = 1;break;
	   }
	}
	else {
	   printf("Invalid character %c in ADDEFATT environment variable.\n",
		   character);
	   exit(1);
	}
     }
  }
  else {
     def_attrib = 0x3f;
     def_zero = 1;
  }

  /* If the first command line parameter is not a switch, then file_parm   */
  /* is assigned the first command line parameter, if it is a switch or	   */
  /* no parameters are passed, then file_parm defualts to ""		   */

  if (parms[1][0] != '/' && num_parms >= 2) {
     file_parm_flag = 1;
     strcpy(file_parm,parms[1]);
     strupr(file_parm);
  }
  else {
     file_parm_flag = 0;
     file_parm[0] = 0;
  }
  /* now we split file_parm */
  pdrive[0] = 0;
  ppath[0] = 0;
  pname[0] = 0;
  pext[0] = 0;
  fnsplit(file_parm,pdrive,ppath,pname,pext);
  /* check for the "." and ".." directories at the end of file_parm	   */
  /* and place in pname and make pext = null string			   */
  if (strcmp(pname,".") == 0 && strcmp(pext,".") == 0) {
     strcpy(pname,"..");  /* if ".." was at the end */
     pext[0] = 0;
  }
  else {
     if (pname[0] == 0 && strcmp(pext,".") == 0) {
	pname[0] = 0;  /* if "." was at the end */
	pext[0] = 0;   /* we just get rid of it, since it is redundant */
     }
  }
  /* we take care of the drive. If it is not given we get the deault drive */
  /* if it is lower case we make it upper case, and drive is the number.   */
  if (pdrive[0] == 0) {
     drive = getdisk();
     pdrive[0] = drive + 'A';
     pdrive[1] = ':';
     pdrive[2] = 0;
  }
  else {
     drive = pdrive[0] - 'A';
  }
  /* If we have no path, set ppath to the current working directory and add */
  /* preceeding & trailing \'s						    */
  if (ppath[0] == 0) {
     strcpy(ppath,"\\");
     getcurdir(drive + 1,&ppath[1]);
     /* add the trailing \ only if needed */
     if (ppath[strlen(ppath)-1] != '\\') {
	strcat(ppath,"\\");
     }
  }

  /* If the user supplied the path, we will check to see if it's valid	   */
  else {
     strcpy(file_parm,pdrive);
     strcat(file_parm,ppath);
     strcat(file_parm,"*.*");
     done = findfirst(file_parm,&file_block,0x3f);
     /* If the path was not found, print a message and exit */
     /* Every path will have at least the filenames "." and ".." */
     if (done == -1 && errno == ENOENT) {
	printf("\nInvalid Path !\n");
	exit(1);
     }
  }
  /* If pname = ".." we will hack one sub-dir off ppath			   */
  if (strcmp(pname,"..") == 0) {
     /* If there is only the root directory or the path does not start with */
     /* a back-slash, then we want to give an error message and leave.	   */
     if (strlen(ppath) < 2 || ppath[0] != '\\') {
	printf("\nInvalid Path !\n");
	exit(1);
     }
     for (index = strlen(ppath) - 1;index > 0;index--) {
	ppath[index] = 0;
	if (ppath[index-1] == '\\') {  /* if we run into the next back-slash */
	   break;		      /* quit before we hack it off	    */
	}
     }
     pname[0] = 0;   /* make pname = null */
  }
  /* If pname and pext is a directory then we want to add it to ppath and  */
  /* set pname to "*" and pext to ".*"					   */

  /* check to see if the pname is a directory	 */

  if (pname[0] != 0 && strcmp(pname,"..") != 0) {
     strcpy(file_parm,pdrive);
     strcat(file_parm,ppath);
     strcat(file_parm,pname);
     strcat(file_parm,pext);
     done = findfirst(file_parm,&file_block,FA_DIREC);
     name[0] = 0;
     ext[0] = 0;
     fnsplit(file_block.ff_name,NULL,NULL,name,ext);

  /* If we found a directory and it has the same file name and extension as */
  /* pname and pext, then we add the name to ppath as well as a trailing    */
  /* \ and we set pname and pext to null.				    */

     if (!done && file_block.ff_attrib & FA_DIREC
	 && strcmp(pname,name) == 0
	 && strcmp(pext,ext) == 0)  {
	strcat(ppath,name);
	strcat(ppath,ext);
	strcat(ppath,"\\");
	pname[0] = 0;
	pext[0] = 0;
     }
  }

  /* If we don't have a file name here, then we make it = "*"		    */

  if (pname[0] == 0) {
     strcpy(pname,"*");
  }

  /* If we don't have an extension here, then we make it = ".*"		    */

  if (pext[0] == 0) {
     strcpy(pext,".*");
  }

  /* Now we convert all lower case to upper case and put the pieces	    */
  /* together to make file_parm, and make proot = pdrive + "\".		    */

  strcpy(file_parm,pdrive);
  strcat(file_parm,ppath);
  strcat(file_parm,pname);
  strcat(file_parm,pext);
  strcpy(proot,pdrive);
  strcat(proot,"\\*.*");

  /* Here we parse out the command line switches, assuming that the first   */
  /* command line paramter may be a switch, and creat a file attribute mask */
  /* to use in our main selection loop. If a /X switch is passed, then we   */
  /* turn on all attribute switches in our mask. If a /0 or /O switch is    */
  /* passed, then set a flag that tells us to include file with no	    */
  /* attribute bits set.						    */
  /* Also, if no command line switches are passed, then all attributes in   */
  /* the mask are set. (The same effect as using /X)			    */

  /* concatenate all command line parms other than the file name	    */

  for (index = 1+file_parm_flag,switches[0] = 0;index < num_parms;index ++) {
     strcat(switches,parms[index]);
  }

  /* Here we count up all attribute selection switches, if any		    */
  for (index = 0,num_switches = 0,len = strlen(switches);index < len;index++) {
     point = strchr("AaDdHhLlRrSsXxOo",switches[index]);
     if (point) {
	num_switches++;
     }
  }
  len = strlen(switches);
  attrib = 0;
  for (index = 0;index < len;index++) {
     character = switches[index];
     point = strchr(" /AaDdHhLlRrSsXxOo0125",character);
     if (point) {
	switch (character) {
	   case 'A' :
	   case 'a' : attrib |= FA_ARCH;break;
	   case 'D' :
	   case 'd' : attrib |= FA_DIREC;break;
	   case 'H' :
	   case 'h' : attrib |= FA_HIDDEN;break;
	   case 'L' :
	   case 'l' : attrib |= FA_LABEL;break;
	   case 'R' :
	   case 'r' : attrib |= FA_RDONLY;break;
	   case 'S' :
	   case 's' : attrib |= FA_SYSTEM;break;
	   case 'X' :
	   case 'x' : attrib = 0x3f;break;
	   case 'O' :
	   case 'o' :
	   case '0' : zero = 1;break;
	   case '1' :
	   case '2' :
	   case '5' : num_cols = character - '0';
	}
     }
     else {
	printf("Invalid character %c in command line switches.\n",character);
	exit(1);
     }
  }
  if (num_switches == 0) {   /* if no attribute selection */
     attrib = def_attrib;   /* are given, then give default*/
     zero = def_zero;	    /* attribute mask		   */
  }

  /* Now we find the Volume label for pdrive and print an appropraite head  */

  printf("\nVolume in Drive %s ",pdrive);
  flag = 0;
  done = findfirst(proot,&file_block,FA_LABEL);
  while (!done) {
     if (file_block.ff_attrib & FA_LABEL) {
	printf("is %s\n",file_block.ff_name);
	flag = 1;
	break;
     }
     done = findnext(&file_block);
  }
  if (!flag) {
     printf("has no label\n");
  }

  /* Now we print out a heading that contains the pdrive and ppath	    */

  printf("\nAttribute-Directory listing of %s%s\n\n",pdrive,ppath);
  switch (num_cols) {
     case 1 :
	printf("filename ext    size      date     time    attributes\n");
	printf("-------- --- ---------- -------- --------- -----------\n");
	break;
     case 2 :
	printf("filename ext   size     date   attrib  filename ext   size     date   attrib\n");
	printf("-------- --- -------- -------- ------  -------- --- -------- -------- ------\n");
	break;
     case 5 :
	break;
  }

  /* Now we enter our main loop, selecting only files with an attribute	    */
  /* byte containing attribute bits selected in our mask. If a 0x00 mask is */
  /* used, then we select only files with no attribute bits set. If all	    */
  /* attribute bits are set in our mask, then we select all files specified */
  /* by file_parm.							    */

  file_count = 0;
  total_size = 0;
  done = findfirst(file_parm,&file_block,0x3f);
  while (!done) {
     if ((file_block.ff_attrib & attrib)
	 || (zero  && file_block.ff_attrib == 0)
	 || (attrib == 0x3f)) {

	/* parse attribute byte into characters for our output line */

	parse_attribs(&file_block);
	strcpy(ext,"    ");

	/* here we split ff_name into name and extension for output */
	/* we won't split file names "." or ".."		    */

	if (strcmp(file_block.ff_name,".") != 0
	    && strcmp(file_block.ff_name,"..") != 0) {
	   fnsplit(file_block.ff_name,NULL,NULL,name,ext);
	}
	else {
	   strcpy(name,file_block.ff_name);
	   ext[0] = 0;
	}

	/* now we format our output line and print it. */
	/* and increment file_count and add the file size to total_size */

	file_count += 1;
	total_size += file_block.ff_fsize;
	switch (num_cols) {
	   case 1:
	      printf("%-8s %-3s %10ld %s %c %c %c %c %c %c\n",
		      name,
		      &ext[1],
		      file_block.ff_fsize,

		/* convert ff_ftime & ff_fdate to a string for our ouput line*/

		      cftime(file_block.ff_ftime,file_block.ff_fdate,
			     date_time_string),

		/* attribute chracters parsed out by parse_attribs */

		      at_rdonly,
		      at_hidden,
		      at_system,
		      at_label,
		      at_direc,
		      at_arch);
	      break;
	   case 2 :
	      printf("%-8s %-3s %8ld %8.8s %c%c%c%c%c%c",
		      name,
		      &ext[1],
		      file_block.ff_fsize,

		/* convert ff_ftime & ff_fdate to a string for our ouput line*/

		      cftime(file_block.ff_ftime,file_block.ff_fdate,
			     date_time_string),

		/* attribute chracters parsed out by parse_attribs */

		      at_rdonly,
		      at_hidden,
		      at_system,
		      at_label,
		      at_direc,
		      at_arch);
	      if (++col_count == 1) {
		 printf("  ");
	      }
	      else {
		 printf("\n");
		 col_count = 0;
	      }
	      break;
	   case 5 :
	      printf("%-12s",file_block.ff_name);
	      if (++col_count < 5) {
		 printf("    ");
	      }
	      else {
		 printf("\n");
		 col_count = 0;
	      }
	      break;
	}
     }
     done = findnext(&file_block);
  }
  switch (num_cols) {
     case 1 :
	break;
     case 2 :
     case 5 :
	if (col_count > 0) {
	   printf("\n");
	}
	break;
  }

  /* display total number of files displayed, the total of their size, and   */
  /* the free space on drive */

  getdfree(drive + 1,&dfree_space);
  free_space = (unsigned long int)dfree_space.df_sclus
	       * (unsigned long int)dfree_space.df_bsec
	       * (unsigned long int)dfree_space.df_avail;
  printf("\n%5d File(s) using %10ld bytes.\n",file_count,total_size);
  printf("       Drive %2s has %10ld bytes free.",pdrive,free_space);

}

void parse_attribs(struct ffblk *f_blk)
{
   at_rdonly = at_hidden = at_system = at_label = at_direc = at_arch = ' ';
   if (f_blk->ff_attrib & FA_RDONLY) {
      at_rdonly = 'R';
   }
   if (f_blk->ff_attrib & FA_HIDDEN) {
      at_hidden = 'H';
   }
   if (f_blk->ff_attrib & FA_SYSTEM) {
      at_system = 'S';
   }
   if (f_blk->ff_attrib & FA_LABEL) {
      at_label = 'L';
   }
   if (f_blk->ff_attrib & FA_DIREC) {
      at_direc = 'D';
   }
   if (f_blk->ff_attrib & FA_ARCH) {
      at_arch = 'A';
   }
}
char *cftime(int f_time, int f_date,char *out_string)
{
   union {
      int f__time;
      struct {
	 unsigned fttsec: 5;
	 unsigned ftmin: 6;
	 unsigned fthour: 5;
      } bit_time;
   } un_time;

   union {
      int f__date;
      struct {
	 unsigned ftday: 5;
	 unsigned ftmonth: 4;
	 unsigned ftyear: 7;
      } bit_date;
   } un_date;
   un_time.f__time = f_time;
   un_date.f__date = f_date;
   sprintf(out_string,"%2d/%02d/%02d %2d:%02d:%02d%c",
	   un_date.bit_date.ftmonth,
	   un_date.bit_date.ftday,
	   un_date.bit_date.ftyear + (un_date.bit_date.ftyear > 19?-20:80),
	   un_time.bit_time.fthour - (un_time.bit_time.fthour > 12?12:0),
	   un_time.bit_time.ftmin,
	   un_time.bit_time.fttsec,
	   (un_time.bit_time.fthour > 12?'p':'a'));
   return(out_string);
}

